#include "efence.h"

#include "dcompmgr.h"
#include "render.h"
#include "screen.h"
#include "window.h"
#include "time.h"
#include "display.h"
#include "list.h"
#include <stdio.h>
#include <assert.h>
#include <stdlib.h>
#include <xcb/render.h>

static int plugin_id;

#define WINDOW_TYPE_FADE(t) (t != DC_WINDOW_TYPE_DESKTOP)

typedef struct {
    void (*window_show)(d_window_t *w);
    void (*window_hide)(d_window_t *w);
    d_list_t *fades;

//    unsigned int fade_step_size;
    unsigned int fade_step_time;
    unsigned int fade_total_time;

    struct timeval next_timeout;
} data_t;

typedef struct {
    d_window_t *w;
    uint16_t    start_alpha;
    uint16_t    end_alpha;
    uint16_t    current_alpha;
    struct timeval start_time;
    struct timeval end_time;
} fade_t;

static void fade_window_show(d_window_t *w);
static void fade_window_hide(d_window_t *w);
static void start_fade(d_window_t *w, uint16_t start_alpha, uint16_t end_alpha,
                       gboolean refalways);

void
fade_init(d_screen_t *sc, int id)
{
    plugin_id = id;

    data_t *d = malloc(sizeof(data_t));
    d->window_show = sc->window_show;
    d->window_hide = sc->window_hide;
    screen_add_plugin_data(sc, plugin_id, d);

    sc->window_show = fade_window_show;
    sc->window_hide = fade_window_hide;

    d->fades = list_new();
    d->fade_step_time = 15000;     /* 15 milliseconds */
    d->fade_total_time = 180000;  /* 0.18 seconds */
}

void
fade_free(d_screen_t *sc)
{
    data_t *d;
    d_list_it_t *it;

    d = screen_find_plugin_data(sc, plugin_id);
    screen_remove_plugin_data(sc, plugin_id);

    if (d) {
        for (it = list_top(d->fades); it; it = it->next) {
            fade_t *fade = it->data;
            window_set_opacity(fade->w, 0xffff);
            window_zombie_unref(fade->w);
            window_unref(fade->w);
            free(fade);
        }
        list_unref(d->fades);
        free(d);
    }
}

static void
fade_window_show(d_window_t *w)
{
    data_t *d;
    d_list_it_t *it;

    d = screen_find_plugin_data(w->sc, plugin_id);

    /* kill the zombies off cuz a new copy is being shown */
    for (it = list_top(d->fades); it; it = it->next) {
        fade_t *fade = it->data;
        if (fade->w == w) {
            window_zombie_unref(w);
            window_unref(w);
            break;
        }
    }

    d->window_show(w);    

    if (!window_is_input_only(w) && dcompmgr_running()) {
        d_window_type_t type = window_get_type(w);
        if (WINDOW_TYPE_FADE(type))
            start_fade(w, 0x2000, 0xffff, TRUE);
    }
}

static void
fade_window_hide(d_window_t *w)
{
    data_t *d;

    if (!window_is_input_only(w) && dcompmgr_running()) {
        d_window_type_t type = window_get_type(w);
        if (WINDOW_TYPE_FADE(type))
            start_fade(w, 0xffff, 0x0002, FALSE);
    }

    d = screen_find_plugin_data(w->sc, plugin_id);
    d->window_hide(w);
}

static void
start_fade(d_window_t *w, uint16_t start_alpha, uint16_t end_alpha,
           gboolean refalways)
{
    data_t *d;
    d_list_it_t *it;
    fade_t *fade;
    gboolean newfade;
    struct timeval now;

    d = screen_find_plugin_data(w->sc, plugin_id);
    gettimeofday(&now, NULL);

    if (list_length(d->fades) == 0) {
        d->next_timeout = now;
        time_add(&d->next_timeout, d->fade_step_time);
    }

    /* look for an existing one */
    fade = NULL;
    for (it = list_top(d->fades); it; it = it->next) {
        fade = it->data;
        if (fade->w == w) break;
    }
    if (!it) {
        fade = malloc(sizeof(fade_t));
        fade->w = w;
        list_append(d->fades, fade);
        newfade = TRUE;
    }
    else
        newfade = FALSE;

    fade->start_alpha = start_alpha;
    fade->end_alpha = end_alpha;

    if (newfade) {
        fade->start_time = now;
        fade->end_time = now;
        time_add(&fade->end_time, d->fade_total_time);
        fade->current_alpha = start_alpha;

        window_set_opacity(w, fade->current_alpha);
    }
    else {
        /* figure out how far we have to go to finish the fade from where
           we are from the previous fade */
        long remain, total;
        double percent;

        total = ABS(fade->end_alpha - fade->start_alpha);
        remain = ABS(fade->current_alpha - fade->end_alpha);
        percent = (double)remain / total;

        //printf("start %d end %d current %d\n", fade->start, fade->end,
        //       fade->current);
        //printf("remain %lu total %lu percent %f\n", remain, total, percent);

        fade->start_time = now;
        time_add(&fade->start_time, -(1-percent)*d->fade_total_time);
        fade->end_time = now;
        time_add(&fade->end_time, percent*d->fade_total_time);
    }
    if (newfade || refalways) {
        window_zombie_ref(w);
        window_ref(w);
    }
}

int
fade_next_timeout(struct d_screen *sc, struct timeval *tv)
{
    data_t *d;

    d = screen_find_plugin_data(sc, plugin_id);

    if (!d || list_length(d->fades) == 0)
        return FALSE;
    else {
        *tv = d->next_timeout;
        return TRUE;
    }
}

void
fade_timeout(struct d_screen *sc, const struct timeval *now)
{
    data_t *d;
    d_list_it_t *it, *next;

    d = screen_find_plugin_data(sc, plugin_id);

    for (it = list_top(d->fades); it; it = next) {
        fade_t *fade = it->data;

        next = it->next;

        if (time_compare(&fade->end_time, now) <= 0) {
            /* done */
            fade->current_alpha = fade->end_alpha;
        }
        else {
            struct timeval total_time, time_left;
            unsigned long remain, total;
            double percent;

            time_difference(&fade->end_time, now, &time_left);
            time_difference(&fade->end_time, &fade->start_time, &total_time);


            /* count milliseconds */
            remain = time_left.tv_sec * 1000 + time_left.tv_usec / 1000;
            total = total_time.tv_sec * 1000 + total_time.tv_usec / 1000;
            percent = (double)remain / total;

            //printf("done %lu total %lu percent %f\n", done, total, percent);

            if (fade->end_alpha > fade->start_alpha)
                /* increasing */
                fade->current_alpha = fade->end_alpha -
                    (fade->end_alpha - fade->start_alpha) * percent;
            else
                /* decreasing */
                fade->current_alpha = fade->end_alpha +
                    (fade->start_alpha - fade->end_alpha) * percent;
        }

        window_set_opacity(fade->w, fade->current_alpha);

        if (fade->current_alpha == fade->end_alpha) {
            list_delete_link(d->fades, it);
            window_zombie_unref(fade->w);
            window_unref(fade->w);

            g_free(fade);
        }
    }
    d->next_timeout = *now;
    time_add(&d->next_timeout, d->fade_step_time);
}
