#include "efence.h"

#include "render.h"
#include "screen.h"
#include "window.h"
#include "display.h"
#include "list.h"
#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <stdlib.h>

#include <X11/Xmd.h>
#include <GL/glxproto.h>

#include <GL/gl.h>
#include <GL/glx.h>
#include <GL/glxtokens.h>
#include <GL/glext.h>
#include <GL/glxext.h>

#define MAX_DEPTH 32

static int plugin_id;

#define LEFT   0
#define TOP    1
#define RIGHT  2
#define BOTTOM 3

typedef void (*BindEXTFunc)(Display *, GLXDrawable, int, const int *);
typedef void (*ReleaseEXTFunc)(Display *, GLXDrawable, int);
typedef int (*SwapIntervalSGIFunc)(int);

typedef struct {
    void (*screen_paint)(d_screen_t *sc);
    void (*screen_root_pixmap_change)(d_screen_t *sc);
    void (*window_show)(d_window_t *w);
    void (*window_zombie_dead)(d_window_t *w);
    void (*window_resize)(d_window_t *w);
    void (*window_reshape)(d_window_t *w);

    float shadowalpha;
    int xshadowoff;
    int yshadowoff;

    GLXFBConfig fbconfig[MAX_DEPTH + 1];

    GLXContext context;
    GLuint root_texname;
    GLXPixmap root_glpixmap;

    BindEXTFunc bind_func;
    ReleaseEXTFunc release_func;
} data_t;

typedef struct {
    GLuint texname;
    GLXPixmap glpixmap;

    GLfloat *texcoords;
    GLint *vertices;
    int nrects;

    xcb_xfixes_fetch_region_cookie_t ck_region;
    int waiting_region;
    int x_region;
    int y_region;
    int w_region;
    int h_region;
    int bw_region;
} window_data_t;

static gboolean glxrender_find_fb_config(d_screen_t *sc, data_t *d);

static void glxrender_paint(d_screen_t *sc);
static void glxrender_root_pixmap_change(d_screen_t *sc);
static void paint_root(d_screen_t *sc, data_t *d);
static void paint_window(d_window_t *window, data_t *d, window_data_t *wd,
                         int x, int y, GLfloat z);
static void paint_shadow(data_t *d, window_data_t *wd, int x, int y,
                         GLfloat z);
static void glxrender_update_window_pixmap(d_window_t *w, data_t *d,
                                           window_data_t *wd);
static void glxrender_fetch_window_region(d_window_t *w, window_data_t *wd);
static void glxrender_update_window_region(d_window_t *w, window_data_t *wd);
static void glxrender_free_window_pixmap(d_window_t *w, data_t *d,
                                         window_data_t *wd);
static void glxrender_free_window_region(window_data_t *wd);
static void glxrender_free_root_pixmap(d_screen_t *sc, data_t *d);
static void glxrender_update_root_pixmap(d_screen_t *sc, data_t *d);

static void glxrender_window_show(d_window_t *window);
static void glxrender_window_zombie_dead(d_window_t *window);
static void glxrender_window_resize(d_window_t *window);
static void glxrender_window_reshape(d_window_t *window);

void
glxrender_init(d_screen_t *sc, int id)
{
    static int context_visual_config[] = {
        GLX_DEPTH_SIZE, 1,
        GLX_DOUBLEBUFFER,
        GLX_RGBA,
        XCB_NONE
    };
    XVisualInfo *vi;
    data_t *d;

    plugin_id = id;

    d = malloc(sizeof(data_t));
    d->screen_paint = sc->screen_paint;
    d->screen_root_pixmap_change = sc->screen_root_pixmap_change;
    d->window_show = sc->window_show;
    d->window_zombie_dead = sc->window_zombie_dead;
    d->window_resize = sc->window_resize;
    d->window_reshape = sc->window_reshape;
    screen_add_plugin_data(sc, plugin_id, d);

    sc->screen_paint = glxrender_paint;
    sc->screen_root_pixmap_change = glxrender_root_pixmap_change;
    sc->window_show = glxrender_window_show;
    sc->window_zombie_dead = glxrender_window_zombie_dead;
    sc->window_resize = glxrender_window_resize;
    sc->window_reshape = glxrender_window_reshape;

    d->shadowalpha = 0.2f; /* 20% */
    d->xshadowoff = 2;
    d->yshadowoff = 2;

    vi = glXChooseVisual(sc->dpy->xlib_dpy, sc->num, context_visual_config);
    if (!vi) {
        printf("unable to find a valid double buffered GL context to use\n");
        exit(1);
    }

    if (!(glxrender_find_fb_config(sc, d))) {
        printf("unable to find FB configs\n");
        exit(1);
    }

    d->context = glXCreateContext(sc->dpy->xlib_dpy, vi, NULL, GL_TRUE);
    glXMakeCurrent(sc->dpy->xlib_dpy, sc->overlay, d->context);

    glViewport(0, 0, sc->super->width_in_pixels, sc->super->height_in_pixels);
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glOrtho(0, sc->super->width_in_pixels, sc->super->height_in_pixels, 0,
            -100.0, 100.0);
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    glEnable(GL_TEXTURE_2D);
    glEnable(GL_DEPTH_TEST);
    //glEnable(GL_SCISSOR_TEST); ?? what is this, glxcompmgr enables it
    //glEnable(GL_STENCIL_TEST); ?? what is this, glxcompmgr enables it

    glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
    glBlendFunc(GL_ONE, GL_ONE_MINUS_SRC_ALPHA);
    glDisable(GL_BLEND);

    glClearColor(0.0f, 0.0f, 0.0f, 1.0f);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    glXSwapBuffers(sc->dpy->xlib_dpy, sc->overlay);

    d->bind_func = (BindEXTFunc)
        glXGetProcAddress((const guchar*)"glXBindTexImageEXT");
    d->release_func = (ReleaseEXTFunc)
        glXGetProcAddress((const guchar*)"glXReleaseTexImageEXT");

    SwapIntervalSGIFunc swap_interval_func = (SwapIntervalSGIFunc)
        glXGetProcAddress((const guchar*)"glXSwapIntervalSGI");
    if(swap_interval_func)
        swap_interval_func(1);

    glGenTextures(1, &d->root_texname);
    d->root_glpixmap = XCB_NONE;
}

static gboolean
glxrender_find_fb_config(d_screen_t *sc, data_t *d)
{
    static const int drawable_tfp_attrs[] = {
        GLX_CONFIG_CAVEAT, GLX_NONE,
        GLX_DOUBLEBUFFER, FALSE,
        GLX_DEPTH_SIZE, 0,
        GLX_RED_SIZE, 1,
        GLX_GREEN_SIZE, 1,
        GLX_BLUE_SIZE, 1,
        GLX_ALPHA_SIZE, 1,
        GLX_RENDER_TYPE, GLX_RGBA_BIT,
        GLX_BIND_TO_TEXTURE_RGBA_EXT, TRUE, /* For TextureFromPixmap */
        XCB_NONE
    };
    int db, stencil, depth, numfb, i;
    GLXFBConfig *fbcons;
    XVisualInfo tvis, *visinfo;

    fbcons = glXChooseFBConfig(sc->dpy->xlib_dpy, sc->num,
                               drawable_tfp_attrs, &numfb);
    if (!fbcons) return FALSE;

    for (i = 0; i <= MAX_DEPTH; i++) {
        int j, count, value;
        VisualID vid;

        vid = 0;
        d->fbconfig[i] = 0;
        db = 32767;
        stencil = 32767;
        depth = 32767;

        //printf("looking for depth %d\n", i);

        tvis.depth = i;
        visinfo = XGetVisualInfo(sc->dpy->xlib_dpy, VisualDepthMask,
                                 &tvis, &count);
        /* pick the nicest visual for the depth */
        for (j = 0; j < count; j++) {
            glXGetConfig(sc->dpy->xlib_dpy, &visinfo[j], GLX_USE_GL, &value);
            if (!value)
                continue;

            glXGetConfig(sc->dpy->xlib_dpy, &visinfo[j], GLX_DOUBLEBUFFER,
                         &value);
            if (value > db)
                continue;
            db = value;

            glXGetConfig(sc->dpy->xlib_dpy, &visinfo[j], GLX_STENCIL_SIZE,
                         &value);
            if (value > stencil)
                continue;
            stencil = value;

            glXGetConfig(sc->dpy->xlib_dpy, &visinfo[j], GLX_DEPTH_SIZE,
                         &value);
            if (value > depth)
                continue;
            depth = value;

            /* use this visual */
            vid = visinfo[j].visualid;
        }

        if (!vid) continue;

        /* look for an fbconfig for this visual */
        for (j = 0; j < numfb; ++j) {
            glXGetFBConfigAttrib(sc->dpy->xlib_dpy, fbcons[j],
                                 GLX_VISUAL_ID, &value);
            if ((unsigned)value == vid) {
                d->fbconfig[i] = fbcons[j];

                printf("found visual 0x%x fbconfig 0x%x for depth %d\n",
                       (uint32_t)vid, (uint32_t)fbcons[j], i);
                break;
            }
        }
    }

    XFree(fbcons);
    return TRUE;
}

void
glxrender_free(d_screen_t *sc)
{
    data_t *d = screen_find_plugin_data(sc, plugin_id);
    if (d) {
        glxrender_free_root_pixmap(sc, d);
        glXDestroyContext(sc->dpy->xlib_dpy, d->context);
        free(d);
    }
    screen_remove_plugin_data(sc, plugin_id);
}

int
glxrender_next_timeout(struct d_screen *sc, struct timeval *tv)
{
    (void)sc;
    (void)tv;
    return FALSE;
}

void
glxrender_timeout(struct d_screen *sc, const struct timeval *now)
{
    (void)sc; (void)now;
}

void
glxrender_window_free_data(d_window_t *w, data_t *d, window_data_t *wd)
{
    glxrender_free_window_pixmap(w, d, wd);
    glxrender_free_window_region(wd);
    glDeleteTextures(1, &wd->texname);
    free(wd);
}

static void
glxrender_window_show(d_window_t *w)
{
    data_t *d;
    window_data_t *wd;

    d = screen_find_plugin_data(w->sc, plugin_id);

    /* pass it on */
    d->window_show(w);

    wd = window_find_plugin_data(w, plugin_id);
    if (wd) {
        glxrender_window_free_data(w, d, wd);
        window_remove_plugin_data(w, plugin_id);
    }

    wd = malloc(sizeof(window_data_t));
    glGenTextures(1, &wd->texname);
    wd->glpixmap = XCB_NONE;
    wd->texcoords = NULL;
    wd->vertices = NULL;
    wd->nrects = 0;
    wd->waiting_region = FALSE;

    window_add_plugin_data(w, plugin_id, wd);

    glxrender_fetch_window_region(w, wd);
}

static void
glxrender_window_zombie_dead(d_window_t *w)
{
    data_t *d;
    window_data_t *wd;

    d = screen_find_plugin_data(w->sc, plugin_id);
    wd = window_find_plugin_data(w, plugin_id);
    if (wd) {
        glxrender_window_free_data(w, d, wd);
        window_remove_plugin_data(w, plugin_id);
    }

    /* pass it on */
    d->window_zombie_dead(w);
}

static void
glxrender_free_window_pixmap(d_window_t *w, data_t *d, window_data_t *wd)
{
    /* this might cause an error, oh well */
    if (wd->glpixmap) {
        glBindTexture(GL_TEXTURE_2D, wd->texname);
        d->release_func(w->sc->dpy->xlib_dpy,
                        wd->glpixmap, GLX_FRONT_LEFT_EXT);
        glBindTexture(GL_TEXTURE_2D, 0);

        glXDestroyPixmap(w->sc->dpy->xlib_dpy, wd->glpixmap);
        wd->glpixmap = XCB_NONE;

    }
}

static void
glxrender_free_window_region(window_data_t *wd)
{
    if (wd->nrects) {
        free(wd->texcoords);
        free(wd->vertices);
        wd->texcoords = NULL;
        wd->vertices = NULL;
        wd->nrects = 0;
    }
}

static void
glxrender_free_root_pixmap(d_screen_t *sc, data_t *d)
{
    /* this might cause an error, oh well */
    if (d->root_glpixmap) {
        glBindTexture(GL_TEXTURE_2D, d->root_texname);
        d->release_func(sc->dpy->xlib_dpy,
                        d->root_glpixmap, GLX_FRONT_LEFT_EXT);
        glBindTexture(GL_TEXTURE_2D, 0);

        glXDestroyPixmap(sc->dpy->xlib_dpy, d->root_glpixmap);
        d->root_glpixmap = XCB_NONE;
    }
}

static void
glxrender_update_window_pixmap(d_window_t *w, data_t *d, window_data_t *wd)
{
    xcb_pixmap_t px;
    uint8_t depth;

    static int attrs[] = {
        GLX_TEXTURE_FORMAT_EXT,
        XCB_NONE,
        XCB_NONE
    };

    px = window_get_pixmap(w);
    depth = window_get_depth(w);

    if (!px) return;

    if (!d->fbconfig[depth]) {
        printf("no GL visual for depth %d\n", depth);
        return;
    }

    if (window_is_argb(w))
        attrs[1] = GLX_TEXTURE_FORMAT_RGBA_EXT;
    else
        attrs[1] = GLX_TEXTURE_FORMAT_RGB_EXT;

    wd->glpixmap = glXCreatePixmap(w->sc->dpy->xlib_dpy,
                                   d->fbconfig[depth],
                                   px, attrs);
#if 0
    wd->glpixmap = xcb_generate_id(w->sc->dpy->conn);
    int v;
    glXGetFBConfigAttrib(w->sc->dpy->xlib_dpy, d->fbconfig[depth],
                         GLX_FBCONFIG_ID, &v);
    xcb_void_cookie_t ck =
        xcb_glx_create_pixmap_checked(w->sc->dpy->conn, w->sc->num,
                                      v,
                                      px, wd->glpixmap,
                                      2,
                                      attrs);
    xcb_generic_error_t *err = xcb_request_check(w->sc->dpy->conn, ck);
    if (err) {
        display_error(w->sc->dpy, err);
        free(err);
        wd->glpixmap = XCB_NONE;
        return;
    }
#endif

    glBindTexture(GL_TEXTURE_2D, wd->texname);
    d->bind_func(w->sc->dpy->xlib_dpy,
                 wd->glpixmap, GLX_FRONT_LEFT_EXT, NULL);
#if 0
    {
        /*
          BindTexImageEXT
          1           CARD8                   opcode (X assigned)
          1           16                      GLX opcode (glXVendorPrivate)
          2           6+n                     request length
          4           1330                    vendor specific opcode
          4           CARD32                  context tag
          4           GLX_DRAWABLE            drawable
          4           INT32                   buffer
          4           CARD32                  num_attributes
          4*n         LISTofATTRIBUTE_PAIR    attribute, value pairs.
        */
        unsigned int len = 12;
        uint32_t data[] = {
            wd->glpixmap,
            GLX_FRONT_LEFT_EXT,
            0,
        };
        xcb_glx_vendor_private(w->sc->dpy->conn,
                               X_GLXvop_BindTexImageEXT,
                               1,
                               len, (uint8_t*)data);
    }
#endif

    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);

    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);

    glBindTexture(GL_TEXTURE_2D, 0);
}

static void
glxrender_fetch_window_region(d_window_t *w, window_data_t *wd)
{
    window_get_area(w, &wd->x_region, &wd->y_region,
                    &wd->w_region, &wd->h_region, &wd->bw_region);

    wd->ck_region =
        xcb_xfixes_fetch_region_unchecked(w->sc->dpy->conn,
                                          window_get_region(w));
    wd->waiting_region = TRUE;
    xcb_flush(w->sc->dpy->conn);
}

static void
glxrender_update_window_region(d_window_t *w, window_data_t *wd)
{
    xcb_xfixes_fetch_region_reply_t *rep;
    xcb_rectangle_t area, *rects;
    int nrects, i, j;

    if (!wd->waiting_region) return;

    area.x = wd->x_region;
    area.y = wd->y_region;
    area.width = wd->w_region + wd->bw_region * 2;
    area.height = wd->h_region + wd->bw_region * 2;

    rep = xcb_xfixes_fetch_region_reply(w->sc->dpy->conn, wd->ck_region, NULL);
    if (!rep) {
        rects = &area;
        nrects = 1;
    }
    else {
        rects = xcb_xfixes_fetch_region_rectangles(rep);
        nrects = xcb_xfixes_fetch_region_rectangles_length(rep);
    }

    wd->texcoords = (GLfloat*)realloc(wd->texcoords,
                                      sizeof(GLfloat) * (nrects * 4));
    wd->vertices = (GLint*)realloc(wd->vertices, sizeof(GLint) * (nrects * 4));
    wd->nrects = nrects;

    for (i = j = 0; i < nrects * 4; ++j, i += 4) {
        wd->texcoords[i+LEFT] =
            (GLfloat)(rects[j].x - area.x) / (GLfloat)area.width;
        wd->texcoords[i+TOP] =
            (GLfloat)(rects[j].y - area.y) / (GLfloat)area.height;
        wd->texcoords[i+RIGHT] =
            (GLfloat)(rects[j].x - area.x + rects[j].width) /
            (GLfloat)area.width;
        wd->texcoords[i+BOTTOM] =
            (GLfloat)(rects[j].y - area.y + rects[j].height) /
            (GLfloat)area.height;

        wd->vertices[i+LEFT] = rects[j].x - area.x;
        wd->vertices[i+TOP] = rects[j].y - area.y;
        wd->vertices[i+RIGHT] = rects[j].x - area.x + rects[j].width;
        wd->vertices[i+BOTTOM] = rects[j].y - area.y + rects[j].height;
    }

    if (rep)
        free(rep);

    wd->waiting_region = FALSE;
}

static void
glxrender_update_root_pixmap(d_screen_t *sc, data_t *d)
{
    xcb_pixmap_t px;
    static int attrs[] = {
        GLX_TEXTURE_FORMAT_EXT,
        GLX_TEXTURE_FORMAT_RGB_EXT,
        XCB_NONE
    };

    px = screen_get_root_pixmap(sc);
    if (!px) return;

    if (!d->fbconfig[sc->super->root_depth]) {
        printf("no GL visual for depth %d\n", sc->super->root_depth);
        return;
    }

    d->root_glpixmap = glXCreatePixmap(sc->dpy->xlib_dpy,
                                       d->fbconfig[sc->super->root_depth],
                                       px, attrs);


    glBindTexture(GL_TEXTURE_2D, d->root_texname);
    d->bind_func(sc->dpy->xlib_dpy,
                 d->root_glpixmap, GLX_FRONT_LEFT_EXT, NULL);

    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);

    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);

    glBindTexture(GL_TEXTURE_2D, 0);
}

static void
glxrender_window_resize(d_window_t *w)
{
    data_t *d;
    window_data_t *wd;

    d = screen_find_plugin_data(w->sc, plugin_id);
    wd = window_find_plugin_data(w, plugin_id);

    /* pass it on */
    d->window_resize(w);

    assert(wd != NULL);
    glxrender_free_window_pixmap(w, d, wd);
    glxrender_fetch_window_region(w, wd);
}

static void
glxrender_window_reshape(d_window_t *w)
{
    data_t *d;
    window_data_t *wd;

    d = screen_find_plugin_data(w->sc, plugin_id);
    wd = window_find_plugin_data(w, plugin_id);

    /* pass it on */
    d->window_reshape(w);

    assert(wd != NULL);
    //glxrender_free_window_pixmap(w, d, wd);
    glxrender_fetch_window_region(w, wd);
}

static void
glxrender_root_pixmap_change(d_screen_t *sc)
{
    data_t *d;

    d = screen_find_plugin_data(sc, plugin_id);
    glxrender_free_root_pixmap(sc, d);

    /* pass it on */
    d->screen_root_pixmap_change(sc);
}

static void
glxrender_paint(d_screen_t *sc)
{
    data_t *d = screen_find_plugin_data(sc, plugin_id);
    d_list_it_t *it;
    GLfloat z;

    //printf("painting\n");

    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    /* add 0.1f to keep everything above the root */
    z = list_length(sc->stacking) * 0.1f + 0.1f;
    for (it = list_top(sc->stacking); it; it = it->next) {
        d_window_t *w = it->data;

        if (!window_is_input_only(w) &&
            (window_is_mapped(w) || window_is_zombie(w)))
        {
            int x, y, width, height, bwidth;
            gboolean opaque;
            window_data_t *wd;

            window_get_area(w, &x, &y, &width, &height, &bwidth);

            if (!(x < sc->super->width_in_pixels &&
                  y < sc->super->height_in_pixels &&
                  (x + width > 0 || x + width + d->xshadowoff > 0) &&
                  (y + height > 0 || y + height + d->yshadowoff > 0)))
            {
                continue;
            }

            opaque = !window_is_argb(w) && window_get_opacity(w) == 0xffff;

            if (opaque) {
                wd = window_find_plugin_data(w, plugin_id);

                //glPushMatrix();

                paint_window(w, d, wd, x, y, z - 0.05f);

                //glPopMatrix();
            }

            z -= 0.1f;
        }
    }

    paint_root(sc, d);

    glEnable(GL_BLEND);
    glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

    for (it = list_bottom(sc->stacking); it; it = it->prev) {
        d_window_t *w = it->data;

        if (!window_is_input_only(w) &&
            (window_is_mapped(w) || window_is_zombie(w)))
        {
            int x, y, width, height, bwidth;
            gboolean opaque;
            window_data_t *wd;
            uint16_t opac;
            GLfloat alpha;

            window_get_area(w, &x, &y, &width, &height, &bwidth);

            if (!(x < sc->super->width_in_pixels &&
                  y < sc->super->height_in_pixels &&
                  (x + width > 0 || x + width + d->xshadowoff > 0) &&
                  (y + height > 0 || y + height + d->yshadowoff > 0)))
            {
                continue;
            }

            opac = window_get_opacity(w);
            opaque = !window_is_argb(w) && opac == 0xffff;

            wd = window_find_plugin_data(w, plugin_id);

            /* black shadow */
            alpha = d->shadowalpha;
            alpha *= opac;
            alpha /= 0xffff;
            if (alpha >= 0.01) {
                glColor4f(0.0f, 0.0f, 0.0f, alpha);
                paint_shadow(d, wd, x, y, z);
            }

            if (!opaque) {
                glColor4us(opac, opac, opac, opac);
                paint_window(w, d, wd, x, y, z + 0.05f);
            }

            z += 0.1f;
        }
    }

    glColor4f(0.0f, 0.0f, 0.0f, 0.0f);
    glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
    glDisable(GL_BLEND);

    glXSwapBuffers(sc->dpy->xlib_dpy, sc->overlay);

    glFinish();

    /* call the function we replaced in the chain */
    d->screen_paint(sc);
}

static void
paint_root(d_screen_t *sc, data_t *d)
{
    if (!d->root_glpixmap)
        glxrender_update_root_pixmap(sc, d);

    glBindTexture(GL_TEXTURE_2D, d->root_texname);
    glBegin(GL_QUADS);
    glTexCoord2f(0, 0);
    glVertex3i(0, 0, 0);
    glTexCoord2f(1, 0);
    glVertex3i(sc->super->width_in_pixels, 0, 0);
    glTexCoord2f(1, 1);
    glVertex3i(sc->super->width_in_pixels, sc->super->height_in_pixels, 0);
    glTexCoord2f(0, 1);
    glVertex3i(0, sc->super->height_in_pixels, 0);
    glEnd();

    glBindTexture(GL_TEXTURE_2D, 0);
}

static void
paint_window(d_window_t *w, data_t *d, window_data_t *wd, int x, int y,
             GLfloat z)
{
    int i;

    if (!wd->glpixmap)
        glxrender_update_window_pixmap(w, d, wd);
    if (wd->waiting_region)
        glxrender_update_window_region(w, wd);

    if (wd->nrects < 1) return;

    glBindTexture(GL_TEXTURE_2D, wd->texname);

    glBegin(GL_QUADS);
    for (i = 0; i < wd->nrects * 4; i += 4) {
        glTexCoord2f(wd->texcoords[i+LEFT], wd->texcoords[i+TOP]);
        glVertex3f(x+wd->vertices[i+LEFT], y+wd->vertices[i+TOP], z);
        glTexCoord2f(wd->texcoords[i+RIGHT], wd->texcoords[i+TOP]);
        glVertex3f(x+wd->vertices[i+RIGHT], y+wd->vertices[i+TOP], z);
        glTexCoord2f(wd->texcoords[i+RIGHT], wd->texcoords[i+BOTTOM]);
        glVertex3f(x+wd->vertices[i+RIGHT], y+wd->vertices[i+BOTTOM], z);
        glTexCoord2f(wd->texcoords[i+LEFT], wd->texcoords[i+BOTTOM]);
        glVertex3f(x+wd->vertices[i+LEFT], y+wd->vertices[i+BOTTOM], z);
    }
    glEnd();

    glBindTexture(GL_TEXTURE_2D, 0);
}

static void
paint_shadow(data_t *d, window_data_t *wd, int x, int y, GLfloat z)
{
    int i;

    if (wd->nrects < 1) return;

    /* shape the shadow to the window */
    glBindTexture(GL_TEXTURE_2D, wd->texname);

    x += d->xshadowoff;
    y += d->yshadowoff;

    glBegin(GL_QUADS);
    for (i = 0; i < wd->nrects * 4; i += 4) {
        glTexCoord2f(wd->texcoords[i+LEFT], wd->texcoords[i+TOP]);
        glVertex3f(x+wd->vertices[i+LEFT], y+wd->vertices[i+TOP], z);
        glTexCoord2f(wd->texcoords[i+RIGHT], wd->texcoords[i+TOP]);
        glVertex3f(x+wd->vertices[i+RIGHT], y+wd->vertices[i+TOP], z);
        glTexCoord2f(wd->texcoords[i+RIGHT], wd->texcoords[i+BOTTOM]);
        glVertex3f(x+wd->vertices[i+RIGHT], y+wd->vertices[i+BOTTOM], z);
        glTexCoord2f(wd->texcoords[i+LEFT], wd->texcoords[i+BOTTOM]);
        glVertex3f(x+wd->vertices[i+LEFT], y+wd->vertices[i+BOTTOM], z);
    }
    glEnd();

    glBindTexture(GL_TEXTURE_2D, 0);
}
